/*
 *  ========================================================================
 *  DIScrete event baSed Energy Consumption simulaTor 
 *    					             for Clouds and Federations (DISSECT-CF)
 *  ========================================================================
 *  
 *  This file is part of DISSECT-CF.
 *  
 *  DISSECT-CF is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *  
 *  DISSECT-CF is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 *  General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with DISSECT-CF.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *  (C) Copyright 2013, Gabor Kecskemeti (gkecskem@dps.uibk.ac.at,
 *   									  kecskemeti.gabor@sztaki.mta.hu)
 */


package hu.mta.sztaki.lpds.cloud.simulator.storage;

import hu.mta.sztaki.lpds.cloud.simulator.util.FastArray;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;

public class Repository extends NetworkNode {

	final HashMap<String, StorageObject> contents = new HashMap<String, StorageObject>(8192);

	final FastArray<StorageObject> underTransfer = new FastArray<StorageObject>(
			new StorageObject[50]);

	final private long maxStorageCapacity;
	private long currentStorageUse = 0;
	private long promisedStorage = 0;


	public Repository(final long capacity, final long maxInBW,
			final long maxOutBW, final long diskBW) {
		super(maxInBW, maxOutBW, diskBW);
		maxStorageCapacity = capacity;
	}

	public boolean registerObject(final StorageObject so) {
		if (!contents.containsKey(so.id)) {
			final long futureUse = currentStorageUse + promisedStorage == 0 ? so.size
					: (promisedStorage - so.size);
			if (futureUse > maxStorageCapacity) {
				return false;
			}
			contents.put(so.id, so);
			currentStorageUse = currentStorageUse + so.size;
		}
		return true;
	}

	public boolean deregisterObject(final StorageObject so) {
		if (so == null) {
			return true;
		}
		return deregisterObject(so.id);
	}

	public boolean deregisterObject(final String soid) {
		if (soid == null) {
			return true;
		}
		if (0 > underTransfer.iterate(
				new FastArray.ForEachWithFeedback<StorageObject>() {
					@Override
					public boolean eval(StorageObject element) {
						if (element.id.equals(soid)) {
							return true;
						}
						return false;
					}
				}, 0)) {
			contents.remove(soid);
			return true;
		}
		return false;
	}

	public boolean requestContentDelivery(final String id,
			final Repository target, final NetworkNode.NetworkEvent ev) {
		return requestContentDelivery(id, null, target, ev);
	}

	public boolean duplicateContent(final String id, final String newId,
			final NetworkNode.NetworkEvent ev) {
		return requestContentDelivery(id, newId, this, ev);
	}

	public boolean requestContentDelivery(final String id, final String newId,
			final Repository target, final NetworkNode.NetworkEvent ev) {
		if (this == target && (newId == null || newId.equals(id)))
			return false;
		final StorageObject totransfer = contents.get(id);
		if (totransfer == null) {
			return false;
		}
		final Repository myRepo = this;
		final long increasedpromise = target.promisedStorage + totransfer.size;
		if (increasedpromise + target.currentStorageUse <= target.maxStorageCapacity) {
			underTransfer.add(totransfer);
			target.promisedStorage = increasedpromise;
			initTransfer(totransfer.size, myRepo, target,
					new NetworkNode.NetworkEvent() {
						@Override
						public void trComplete() {
							underTransfer.remove(totransfer);
							final StorageObject toRegister = (target == myRepo || newId != null) ? totransfer
									.newCopy(newId) : totransfer;

							if (target.registerObject(toRegister)) {
								target.promisedStorage -= totransfer.size;
								ev.trComplete();
							} else {
								throw new IllegalStateException(
										"THIS SHOULD NOT HAPPEN (there was no"
												+ " space for the stuff to be stored in"
												+ "the repo!!!)");
							}
						}
					});
			return true;
		}
		return false;
	}

	public StorageObject lookup(final String soid) {
		return contents.get(soid);
	}

	public Collection<StorageObject> contents() {
		return Collections.unmodifiableCollection(contents.values());
	}
}
